#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include "fsloc.h"
#include "exportfs.h"

/* Debugging tool: prints out @servers info to syslog */
static void replicas_print(struct servers *sp)
{
	int i;
	if (!sp) {
		syslog(LOG_INFO, "NULL replicas pointer");
		return;
	}
	syslog(LOG_INFO, "replicas listsize=%i", sp->h_num);
	for (i=0; i<sp->h_num; i++) {
		syslog(LOG_INFO, "%s:/%s",
		       sp->h_mp[i]->h_host, sp->h_mp[i]->h_path);
	}
}

/* Called by setting 'Method = stub' in config file.  Just returns
 * some syntactically correct gibberish for testing purposes.
 */
static struct servers *method_stub(char *key)
{
	struct servers *sp;
	struct mount_point *mp;

	syslog(LOG_INFO, "called method_stub");
	sp = malloc(sizeof(struct servers));
	if (!sp)
		return NULL;
	mp = calloc(1, sizeof(struct mount_point));
	if (!mp) {
		free(sp);
		return NULL;
	}
	sp->h_num = 1;
	sp->h_mp[0] = mp;
	mp->h_host = strdup("stub_server");
	mp->h_path = strdup("/my/test/path");
	sp->h_referral = 1;
	return sp;
}

/* Scan @list, which is a NULL-terrminated array of strings of the
 * form host[:host]:/path, and return corresponding servers structure.
 */
static struct servers *parse_list(char **list)
{
	int i;
	struct servers *res;
	struct mount_point *mp;
	char *cp;

	res = malloc(sizeof(struct servers));
	if (!res)
		return NULL;
	res->h_num = 0;

	/* parse each of the answers in sucession. */
	for (i=0; list[i] && i<FSLOC_MAX_LIST; i++) {
		mp = calloc(1, sizeof(struct mount_point));
		if (!mp) {
			release_replicas(res);
			return NULL;
		}
		cp = strstr(list[i], ":/");
		if (!cp) {
			syslog(LOG_WARNING, "invalid entry '%s'", list[i]);
			continue; /* XXX Need better error handling */
		}
		res->h_mp[i] = mp;
		res->h_num++;
		mp->h_host = strndup(list[i], cp - list[i]);
		cp++;
		mp->h_path = strdup(cp);
	}
	return res;
}

/* Converts from path@host[+host][:path@host[+host]] to
 * host[:host]:path[@host[:host]:path]
 *
 * XXX Once the interface is stabilized, we can put the kernel and
 * userland formats into agreement, so this won't be necessary.
 */
static char *strconvert(const char *in)
{
	char *path, *ptr, *copy, *rv, *rvptr, *next;
	next = copy = strdup(in);
	rvptr = rv = malloc(strlen(in) + 1);
	if (!copy || !rv)
		goto error;
	while (next) {
		ptr = strsep(&next, ":");
		path = strsep(&ptr, "@");
		if (!ptr)
			goto error;
		while (*ptr) {
			if (*ptr == '+') {
				*rvptr++ = ':';
				ptr++;
			}
			else
				*rvptr++ =  *ptr++;
		}
		*rvptr++ = ':';
		while (*path) {
			*rvptr++ = *path++;
		}
		if (next)
			*rvptr++ = '@';
		else
			*rvptr = '\0';
	}
	free(copy);
	return rv;
error:
	free(copy);
	free(rv);
	return NULL;
}

/* @data is a string of form path@host[+host][:path@host[+host]]
 */
static struct servers *method_list(char *data)
{
	char *copy, *ptr=data;
	char **list;
	int i, listsize;
	struct servers *rv=NULL;

	syslog(LOG_INFO, "method_list(%s)\n", data);
	for (ptr--, listsize=1; ptr; ptr=index(ptr, ':'), listsize++)
		ptr++;
	list = malloc(listsize * sizeof(char *));
	copy = strconvert(data);
	syslog(LOG_INFO, "converted to %s\n", copy);
	if (list && copy) {
		ptr = copy;
		for (i=0; i<listsize; i++) {
			list[i] = strsep(&ptr, "@");
		}
		rv = parse_list(list);
	}
	free(copy);
	free(list);
	replicas_print(rv);
	return rv;
}

/* Returns appropriately filled struct servers, or NULL if had a problem */
struct servers *replicas_lookup(int method, char *data, char *key)
{
	struct servers *sp=NULL;
	switch(method) {
	case FSLOC_NONE:
		break;
	case FSLOC_REFER:
		sp = method_list(data);
		if (sp)
			sp->h_referral = 1;
		break;
	case FSLOC_REPLICA:
		sp = method_list(data);
		if (sp)
			sp->h_referral = 0;
		break;
	case FSLOC_STUB:
		sp = method_stub(data);
		break;
	default:
		syslog(LOG_WARNING, "Unknown method = %i", method);
	}
	replicas_print(sp);
	return sp;
}

void release_replicas(struct servers *server)
{
	int i;

	if (!server) return;
	for (i = 0; i < server->h_num; i++) {
		free(server->h_mp[i]->h_host);
		free(server->h_mp[i]->h_path);
		free(server->h_mp[i]);
	}
	free(server);
}
